from sonusai.mixture import get_default_config


def doc_seed() -> str:
    default = f"\nDefault value: {get_default_config()['seed']}"
    return """
'seed' is a mixture database configuration parameter that sets the random number
generator seed.
""" + default


def doc_feature() -> str:
    default = f"\nDefault value: {get_default_config()['feature']}"
    return """
'feature' is a mixture database configuration parameter that sets the feature
to use.
""" + default


def doc_target_level_type() -> str:
    default = f"\nDefault value: {get_default_config()['target_level_type']}"
    return """
'target_level_type' is a mixture database configuration parameter that sets the
algorithm to use to determine target energy level for SNR calculations.
Supported values are:

 default    mean of squares
 speech     ITU-T P.56 active speech level method B
""" + default


def doc_targets() -> str:
    default = f"\nDefault value: {get_default_config()['targets']}"
    return """
'targets' is a mixture database configuration parameter that sets the list of
targets to use.

Required field:

 'name'
                File name. May be one of the following:

  audio         Supported formats are .wav, .mp3, .aif, .flac, and .ogg
  glob          Matches file glob patterns
  .yml          The given YAML file is parsed into the list
  .txt          Each line in the given text file indicates an item which
                may be anything in this list (audio, glob, .yml, or .txt)

Optional fields:

 'truth_settings'
                Local overrides for truth. Contains any or all of the
                following:

  'function'    Name of truth function <str>
  'config'      Truth function config <dict>
  'index'       Truth index <int> or list(<int>)

                'index' indicates which truth fields should be set.
                0 indicates none, 1 is first element in truth output
                vector, 2 2nd element, etc.

                Examples:
                  index = 5       truth in class 5, truth(4, 1)
                  index = [1, 5]  truth in classes 1 and 5, truth([0, 4], 1)

                In mutually-exclusive mode, a frame is expected to only
                belong to one class and thus all probabilities must sum to
                1, and there should be a class for "other" or "none". This
                is effectively truth for a classifier with multichannel
                softmax output. SonusAI will automatically calculate class
                num_classes as 1 - sum(truth(1:num_classes-1). For
                example, a classifier for (dog, cat) must have
                num_classes=3 to include "none" in truth(3).

                For multi-label classification each class is an individual
                probability for that class and any given frame can be
                assigned to multiple classes/labels, i.e., the classes are
                not mutually-exclusive. For example, a NN classifier with
                multichannel sigmoid output. In this case, index could
                also be a vector with multiple class indices. num_classes
                should be set to the number of classes/categories.

  'class_balancing_augmentation'
                Target-specific class balancing augmentation override.
                This target will not use the global class balancing
                augmentation rule, but will use this rule instead for
                class balancing operations. If this rule is specified and
                empty, then this target will not be used for class
                balancing.

  'target_level_type'
                Target-specific override for target_level_type.

Example:

targets:
  - name: data/esc50/ESC-50-master/audio/1-*.wav
    truth_settings:
      function: sed
      config:
        thresholds: [-38, -41, -48]
      index: 2
  - name: target.mp3
    truth_settings:
      function: sed
      config:
        thresholds: [-38, -41, -48]
      index: 5
    class_balancing_augmentation: { }
""" + default


def doc_num_classes() -> str:
    default = f"\nDefault value: {get_default_config()['num_classes']}"
    return """
'num_classes' is a mixture database configuration parameter that sets the number of
classes in this dataset. The number of classes is the total number of parameters
(or classes or labels) in the truth. This controls the size of the truth input to
the model.

Note that the model output 'parameters' dimension is NOT necessarily the same size
as the truth 'num_classes' dimension; there may be multiple truth functions combined
in the truth, e.g., for use in loss function calculations.
""" + default


def doc_class_labels() -> str:
    default = f"\nDefault value: {get_default_config()['class_labels']}"
    return """
'class_labels' is a mixture database configuration parameter that sets class labels
in this dataset.
""" + default


def doc_class_weights_threshold() -> str:
    default = f"\nDefault value: {get_default_config()['class_weights_threshold']}"
    return """
'class_weights_threshold' is a mixture database configuration parameter that sets
the threshold for class weights calculation to quantize truth to binary for counting.

Supports scalar or list:

 scalar     use for all classes
 list       must be of num_classes length
""" + default


def doc_truth_mode() -> str:
    default = f"\nDefault value: {get_default_config()['truth_mode']}"
    return """
'truth_mode' is a mixture database configuration parameter that sets the truth output
mode.

Supported values:

 normal     multi-label (no automatic calculation of other class)
            Set 'num_classes' to the actual number of classification categories.

 mutex      mutually-exclusive
            Set 'num_classes' to the actual number of classification categories plus 1
            to include a "none" category which will be set to 1 - truth (where truth is
            active for only one label) for all frames, to guarantee that the sum of all
            truth outputs is equal to 1. This is required to support softmax()
            neural-net outputs for multi-class classification (single label case).
""" + default


def doc_truth_reduction_function() -> str:
    default = f"\nDefault value: {get_default_config()['truth_reduction_function']}"
    return """
'truth_reduction_function' is a mixture database configuration parameter that set the
truth reduction function. It is used during feature generation to reduce sample-based
truth down to transform frame-based truth. The feature generator further reduces this
down to feature frame-based truth (based on stride and decimation).

Supported values:

 max        Returns the max value in a transform frame
 mean       Returns the mean of a transform frame
 index0     Returns the first value in a transform frame
""" + default


def get_truth_functions() -> str:
    from sonusai.mixture import truth_functions

    functions = [function for function in dir(truth_functions) if not function.startswith('__')]
    text = "\nSupported truth functions:\n\n"
    for function in functions:
        docs = getattr(truth_functions, function).__doc__
        if docs is not None:
            text += f"    {function}\n"
            for doc in docs.splitlines():
                text += f"        {doc}\n"
    return text


def doc_truth_settings() -> str:
    import yaml

    from sonusai.mixture import get_default_config
    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['truth_settings'])}"
    return """
'truth_settings' is a mixture database configuration parameter that sets the truth
generation settings for targets. There is a global 'truth_settings' and there may be
target-specific 'truth_settings'.

A truth setting creates a type of truth and is associated with target file(s).
Target files may have multiple truth settings.
Truth may be non-temporal, per sample, or per feature frame and each value may be scalar
or vector.

Note that there is a difference between transform frames and feature frames: a feature
frame can potentially have a stride dimension (which aggregates multiple transform
frames in a single feature).

There are two notions of truth data: truth_t and truth_f. truth_t is what the truth
functions always generate and it is in the time domain. truth_f, or truth in the feature
domain, is created from truth_t in the following manner. First, the
'truth_reduction_function' specified by the config reduces truth_t into truth data in
the transform frame domain. Then, this transform frame domain truth data is passed
into the feature generator which produces feature frame domain truth data. 

The 'truth_settings' parameter specifies the following:

 'function'     Name of truth function
 'config'       Truth function specific config dictionary
 'index'        Truth index is either a single int or a list of ints

                'index' indicates which truth fields should be set.
                0 indicates none, 1 is the first element in the truth output
                vector, 2 is the 2nd element, etc.

                Examples:
                  index = 5       truth in class 5, truth(4, 1)
                  index = [1, 5]  truth in classes 1 and 5, truth([0, 4], 1)

                In mutually-exclusive mode, a frame is expected to only
                belong to one class and thus all probabilities must sum to
                1, and there should be a class for "other" or "none". This
                is effectively truth for a classifier with multichannel
                softmax output. SonusAI will automatically calculate class
                num_classes as 1 - sum(truth(1:num_classes-1). For
                example, a classifier for (dog, cat) must have
                num_classes=3 to include "none" in truth(3).

                For multi-label classification each class is an individual
                probability for that class and any given frame can be
                assigned to multiple classes/labels, i.e., the classes are
                not mutually-exclusive. For example, a NN classifier with
                multichannel sigmoid output. In this case, index could
                also be a vector with multiple class indices. num_classes
                should be set to the number of classes/categories.
""" + get_truth_functions() + default


def doc_asr_manifest() -> str:
    default = f"\nDefault value: {get_default_config()['asr_manifest']}"
    return """
'asr_manifest' is a mixture database configuration parameter that defines an
optional ASR manifest.

The parameter takes a list of manifest files to be used to populate ASR data
per target. Each line of the manifest should be in the following format:

{"audio_filepath": "/path/to/audio.wav", "text": "the transcription of the utterance", "duration": 23.147}

The audio_filepath field should provide an absolute path to the audio file corresponding
to the utterance. The text field should contain the full transcript for the utterance,
and the duration field should reflect the duration of the utterance in seconds.

Each entry in the manifest (describing one audio file) should be bordered by '{' and '}'
and must be contained on one line. The fields that describe the file should be separated
by commas, and have the form "field_name": value, as shown above.

Since the manifest specifies the path for each utterance, the audio files do not have to be
located in the same directory as the manifest, or even in any specific directory structure.
""" + default


def doc_augmentations() -> str:
    return """
Augmentation Rules

These rules may be specified for target and/or noise. Each rule will be
applied for each target/noise. The values may be specified as scalars, lists,
or random using the syntax: 'rand(<min>, <max>)'.

If a value is specified as a list, then the rule is repeated for each value in
the list.

If a value is specified using rand, then a randomized rule is generated
dynamically per use.

Rules may specify any or all of the following augmentations:

 'normalize'    Normalize audio file to the specified level (in dBFS).
 'gain'         Apply an amplification or an attenuation to the audio signal.
                The signal level is adjusted by the given number of dB; positive
                amplifies, negative attenuates, 0 does nothing.
 'pitch'        Change the audio pitch (but not its tempo). Pitch amount is
                specified as positive or negative 'cents' (i.e., 100ths of a
                semitone).
 'tempo'        Change the audio tempo (but not its pitch). Tempo amount is
                specified as the ratio of the new tempo to the old tempo. For
                example, '1.1' speeds up the tempo by 10% and '0.9' slows it
                down by 10%.
 'eq1'          Apply a two-pole peaking equalization filter. EQ parameters are
                specified as a [frequency, width, gain] triple where:
                  'frequency' gives the central frequency in Hz (20 - SR/2),
                  'width' gives the width as a Q-factor (0.3 - 2.0), and
                  'gain' gives the gain in dB (-20 - 20).
 'eq2'          Apply an additional band of EQ. Same as 'eq1'
 'eq3'          Apply an additional band of EQ. Same as 'eq1'
 'lpf'          Apply a low-pass Butterworth filter. The 3 dB point frequency is
                specified in Hz (20 - SR/2).
 'ir'           An index into a list of impulse responses (specified in the
                'impulse_responses' parameter).
                For targets, the impulse response is applied AFTER truth generation
                and the resulting audio is still aligned with the truth. Random
                syntax for 'ir' is simply 'rand' (i.e., do not specify <min> and <max>).

Only the specified augmentations for a given rule are applied; all others are
skipped in the given rule. For example, if a rule only specifies 'tempo',
then only a tempo augmentation is applied and all other possible augmentations
are ignored (e.g., 'gain', 'pitch', etc.).

Example:

target_augmentations:
  - normalize: -3.5
  - normalize: -3.5
    pitch: [-300, 300]
    tempo: [0.8, 1.2]
    eq1: [[1000, 0.8, 3], [600, 1.0, -4], [800, 0.6, 0]]
  - normalize: -3.5
    pitch: "rand(-300, 300)"
    eq1: ["rand(100, 6000)", "rand(0.6, 1.0)", "rand(-6, 6)"]
    lpf: "rand(1000, 8000)"
  - tempo: "rand(0.9, 1.1)"
    eq1: [["rand(100, 7500)", 0.8, -10], ["rand(100, 7500)", 0.8, 10]]

There are four rules given in this example.

The first rule is simple:
  - normalize: -3.5

This results in just one augmentation being applied to each target:

  normalize: -3.5

The second rule illustrates the use of lists to specify values:
  - normalize: -3.5
    pitch: [-300, 300]
    tempo: [0.8, 1.2]
    eq1: [[1000, 0.8, 3], [600, 1.0, -4], [800, 0.6, 0]]

There are two values given for pitch, two for tempo, and three for EQ. This
rule expands to 2 * 2 * 3 = 12 unique augmentations being applied to each
target:

  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch: -3, tempo: 0.8, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch: -3, tempo: 1.2, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch:  3, tempo: 0.8, eq1: [ 800, 0.6,  0]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [1000, 0.8,  3]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [ 600, 1.0, -4]
  normalize: -3.5, pitch:  3, tempo: 1.2, eq1: [ 800, 0.6,  0]

The third rule shows the use of rand:
  - normalize: -3.5
    pitch: "rand(-300, 300)"
    eq1: ["rand(100, 6000)", "rand(0.6, 1.0)", "rand(-6, 6)"]
    lpf: "rand(1000, 8000)"

This rule is used to create randomized augmentations per use.

The fourth rule demonstrates the use of scalars, lists, and rand:
  - tempo: [0.9, 1, 1.1]
    eq1: [["rand(100, 7500)", 0.8, -10], ["rand(100, 7500)", 0.8, 10]]

This rule expands to 6 unique augmentations being applied to each target
(list of 3 * list of 2). Here is the expansion:

  tempo: 0.9, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 1.0, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 1.1, eq1: ["rand(100, 7500)", 0.8, -10]
  tempo: 0.9, eq1: ["rand(100, 7500)", 0.8, 10]
  tempo: 1.0, eq1: ["rand(100, 7500)", 0.8, 10]
  tempo: 1.1, eq1: ["rand(100, 7500)", 0.8, 10]"""


def doc_target_augmentations() -> str:
    import yaml

    from sonusai.mixture import get_default_config
    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['target_augmentations'])}"
    return """
'target_augmentations' is a mixture database configuration parameter that
specifies a list of augmentation rules to use for each target.

See 'augmentations' for details on augmentation rules.
""" + default


def doc_class_balancing_augmentation() -> str:
    import yaml

    from sonusai.mixture import get_default_config
    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['class_balancing_augmentation'])}"
    return """
'class_balancing_augmentation' is a mixture database configuration parameter
that sets the default augmentation rule to use for generating class balancing
target data. This rule must contain at least one random entry in order to
guarantee unique additional data.

See 'augmentations' for details on augmentation rules.
""" + default


def doc_class_balancing() -> str:
    default = f"\nDefault value: {get_default_config()['class_balancing']}"
    return """
'class_balancing' is a mixture database configuration parameter that
enables/disables class balancing.

Class balancing ensures that each class in a sound classification dataset is
represented equally (i.e., each class has the same number of augmented targets).
This is achieved by creating new class balancing augmentation rules and applying
them to targets in underrepresented classes to create more augmented targets
for those classes.
""" + default


def doc_noises() -> str:
    default = f"\nDefault value: {get_default_config()['class_balancing']}"
    return """
'noises' is a mixture database configuration parameter that sets the list of
noises to use.

Required field:

 'name'
                File name. May be one of the following:

   audio        Supported formats are .wav, .mp3, .aif, .flac, and .ogg
   glob         Matches file glob patterns
   .yml         The given YAML file is parsed into the list
   .txt         Each line in the given text file indicates an item which
                may be anything in this list (audio, glob, .yml, or .txt)
""" + default


def doc_noise_augmentations() -> str:
    import yaml

    from sonusai.mixture import get_default_config
    default = f"\nDefault value:\n\n{yaml.dump(get_default_config()['noise_augmentations'])}"
    return """
'noise_augmentations' is a mixture database configuration parameter that
specifies a list of augmentation rules to use for each noise.

See 'augmentations' for details on augmentation rules.
""" + default


def doc_snrs() -> str:
    default = f"\nDefault value: {get_default_config()['snrs']}"
    return """
'snrs' is a mixture database configuration parameter that specifies a list
of required signal-to-noise ratios (in dB).

All other augmentations are applied to both target and noise and then the
energy levels are measured and the appropriate noise gain calculated to
achieve the desired SNR.

Special values:

 -99    Noise only mixture (no target)
 99     Target only mixture (no noise)
""" + default


def doc_random_snrs() -> str:
    default = f"\nDefault value: {get_default_config()['random_snrs']}"
    return """
'random_snrs' is a mixture database configuration parameter that specifies a
list of random signal-to-noise ratios. The value(s) must be specified as
random using the syntax: 'rand(<min>, <max>)'.

Random SNRs behave slightly differently from regular or ordered SNRs. As with
ordered SNRs, all other augmentations are applied to both target and noise and
then the energy levels are measured and the appropriate noise gain calculated
to achieve the desired SNR. However, unlike ordered SNRs, the desired SNR is
randomized (per the given rule(s)) for each mixture, i.e., previous random
SNRs are not saved and reused.
""" + default


def doc_noise_mix_mode() -> str:
    default = f"\nDefault value: {get_default_config()['noise_mix_mode']}"
    return """
'noise_mix_mode' is a mixture database configuration parameter that sets
how to mix noises with targets.

Supported modes:

 exhaustive          Use every noise/augmentation with every target/augmentation.
 non-exhaustive      Cycle through every target/augmentation without necessarily
                     using all noise/augmentation combinations (reduced data set).
 non-combinatorial   Combine a target/augmentation with a single cut of a
                     noise/augmentation non-exhaustively (each target/augmentation
                     does not use each noise/augmentation). Cut has a random start
                     and loops back to the beginning if the end of a
                     noise/augmentation is reached.
""" + default


def doc_impulse_responses() -> str:
    default = f"\nDefault value: {get_default_config()['impulse_responses']}"
    return """
'impulse_responses' is a mixture database configuration parameter that specifies a
list of impulse response files to use.

See 'augmentations' for details.
""" + default


def doc_spectral_masks() -> str:
    default = f"\nDefault value: {get_default_config()['spectral_masks']}"
    return """
'spectral_masks' is a mixture database configuration parameter that specifies
a list of spectral mask rules.

All other augmentations are applied including SNR and a mixture is generated
and then the spectral mask rules are applied to the resulting mixture feature.

Rules must specify all the following parameters:

 'f_max_width'      Frequency mask maximum width in bins
 'f_num'            Number of frequency masks to apply (set to 0 to apply none)
 't_max_width'      Time mask maximum width in frames
 't_num'            Number of time masks to apply (set to 0 to apply none)
 't_max_percent'    Upper bound on the width of the time mask in percent
""" + default


def doc_config() -> str:
    from sonusai.mixture import VALID_CONFIGS

    text = '\n'
    text += 'The SonusAI database is defined using a config.yml file.\n\n'
    text += 'See the following for details:\n\n'
    for c in VALID_CONFIGS:
        text += f' {c}\n'
    return text



"""sonusai

usage: sonusai [--version] [--help] <command> [<args>...]

The sonusai commands are:
   calc_metric_spenh            Run speech enhancement and analysis
   doc                          Documentation
   evaluate                     Evaluate model performance
   genft                        Generate feature and truth data
   genmix                       Generate mixture and truth data
   genmixdb                     Generate a mixture database
   gentcst                      Generate target configuration from a subdirectory tree
   keras_onnx                   Convert a trained Keras model to ONNX
   keras_predict                Run Keras predict on a trained model
   keras_train                  Train a model using Keras
   lsdb                         List information about a mixture database
   mkmanifest                   Make ASR manifest JSON file
   mkwav                        Make WAV files from a mixture database
   onnx_predict                 Run ONNX predict on a trained model
   plot                         Plot mixture data
   post_spenh_targetf           Run post-processing for speech enhancement targetf data
   torchl_onnx                  Convert a trained Pytorch Lightning model to ONNX
   torchl_predict               Run Lightning predict on a trained model
   torchl_train                 Train a model using Lightning
   tplot                        Plot truth data
   vars                         List custom SonusAI variables

Aaware Sound and Voice Machine Learning Framework. See 'sonusai help <command>'
for more information on a specific command.

"""
from sonusai import logger


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    commands = (
        'calc_metric_spenh',
        'doc',
        'evaluate',
        'genft',
        'genmix',
        'genmixdb',
        'gentcst',
        'keras_onnx',
        'keras_predict',
        'keras_train',
        'lsdb',
        'mkmanifest',
        'mkwav',
        'onnx_predict',
        'plot',
        'post_spenh_targetf',
        'torchl_onnx',
        'torchl_predict',
        'torchl_train',
        'tplot',
        'vars',
    )

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    command = args['<command>']
    argv = args['<args>']

    from subprocess import call

    import sonusai
    from sonusai import SonusAIError

    if command == 'help':
        if not argv:
            exit(call(['sonusai', '-h']))
        elif argv[0] in commands:
            exit(call(['python', f'{sonusai.BASEDIR}/{argv[0]}.py', '-h']))
        else:
            raise SonusAIError(f"{argv[0]} is not a SonusAI command. See 'sonusai help'.")
    elif command in commands:
        exit(call(['python', f'{sonusai.BASEDIR}/{command}.py'] + argv))

    raise SonusAIError(f"{command} is not a SonusAI command. See 'sonusai help'.")


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from .data import PathInfo
from .data import TranscriptData
from .librispeech import collect_librispeech_transcripts
from .librispeech import get_librispeech_manifest_entry
from .vctk_noisy_speech import collect_vctk_noisy_speech_transcripts
from .vctk_noisy_speech import get_vctk_noisy_speech_manifest_entry



from sonusai.utils import ASRResult
from sonusai.utils.asr_functions.data import Data


def fastwhisper(data: Data) -> ASRResult:
    from os import getpid
    from timeit import default_timer as timer

    from faster_whisper import WhisperModel

    from sonusai import SonusAIError

    whisper_model = data.whisper_model
    pid = getpid()
    # print(f'{pid}: Loading model ...')
    retry = 2
    count = 0
    while True:
        try:
            # To pre-download model, first provide whisper_model_name without whisper_model (or =None)
            if whisper_model is None:
                model = WhisperModel(data.whisper_model_name,
                                     device=data.device,
                                     cpu_threads=data.cpu_threads,
                                     compute_type=data.compute_type)
            else:
                model = WhisperModel(whisper_model,
                                     device=data.device,
                                     cpu_threads=data.cpu_threads,
                                     compute_type=data.compute_type,
                                     local_files_only=True)

            # print(f'{pid}: Done Loading, now transcribing ...')
            s_time = timer()
            segments, info = model.transcribe(data.audio, beam_size=int(data.beam_size))
            segments = list(segments)  # The transcription will actually run here.
            e_time = timer()
            elapsed = e_time - s_time
            transcription = "".join(segment.text for segment in segments)
            # print(f'{pid}: Done transcribing.')
            tmp = ASRResult(text=transcription,
                            lang=info.language,
                            lang_prob=info.language_probability,
                            duration=info.duration,
                            num_segments=len(segments),
                            asr_cpu_time=elapsed
                            )
            return tmp
        except Exception as e:
            count += 1
            print(f'{pid}: Warning: fastwhisper exception: {e}')
            if count >= retry:
                raise SonusAIError(f'{pid}: Fastwhisper exception: {e.args}')


"""
Whisper results:
{
  'text': ' The birch canoe slid on the smooth planks.',
  'segments': [
    {
      'id': 0,
      'seek': 0,
      'start': 0.0,
      'end': 2.4,
      'text': ' The birch canoe slid on the smooth planks.',
      'tokens': [
        50363,
        383,
        35122,
        354,
        47434,
        27803,
        319,
        262,
        7209,
        1410,
        591,
        13,
        50483
      ],
      'temperature': 0.0,
      'avg_logprob': -0.4188103675842285,
      'compression_ratio': 0.8571428571428571,
      'no_speech_prob': 0.003438911633566022
    }
  ],
  'language': 'en'
}
"""



#!/usr/bin/env python3
"""aawscd_probwrite

usage: aawscd_probwrite [-h] [-m MACHINE] [-f FRAMES] FILE

options:
   -h, --help
   -m MACHINE, --machine MACHINE    IP address of aawscd platform. [default: localhost].
   -f FRAMES, --frames FRAMES       Number of frames to capture. [default: 10].

aawscd_probwrite connects to an Aaware platform running aawscd and writes the sound classification
probability output to an HDF5 file.

"""
from threading import Condition
from typing import Optional

import numpy as np
from tqdm import tqdm

CLIENT: str = 'aawscd_probwrite'
TOPIC: str = 'aawscd/sc/prob'
DONE: Condition = Condition()
FRAMES: int = 10
FRAME_COUNT: int = 0
DATA: Optional[np.ndarray] = None
PROGRESS: Optional[tqdm] = None


def shutdown(_signum, _frame) -> None:
    global DONE
    with DONE:
        DONE.notify()


def unpack_prob_entry(entry: int) -> tuple[int, int, int]:
    """Decode the packed probability data: [ 16-bit label | 8-bit zone | 8-bit probability ]."""
    data = np.array(entry, dtype=np.uint32)
    label = np.right_shift(data, 16)
    zone = np.bitwise_and(np.right_shift(data, 8), 0xFF)
    value = np.bitwise_and(entry, 0xFF)
    return int(zone), int(label), int(value)


def parse_prob(payload: list) -> np.ndarray:
    """Parse MQTT probability payload from aawscd.

    The 'aawscd/sc/prob' payload is a list of uint32. Each item contains packed data and is
    unpacked/decoded and inserted into an array.
    """

    # First pass: get the zones and labels to determine the size of the array
    zones = set()
    labels = set()
    for entry in payload:
        zone, label, _ = unpack_prob_entry(entry)
        zones.add(zone)
        labels.add(label)

    # Create the array
    prob = np.zeros((len(list(labels)), len(list(zones))), dtype=np.uint8)

    # Second pass: fill in the array probability values.
    for entry in payload:
        zone, label, value = unpack_prob_entry(entry)
        prob[label, zone] = value

    # Return the array
    return prob


def on_message(_client, _userdata, message):
    import paho.mqtt.client as mqtt
    import yaml

    global TOPIC
    if mqtt.topic_matches_sub(TOPIC, message.topic):
        payload = yaml.safe_load(str(message.payload.decode('utf-8')))
        prob = parse_prob(payload['prob'])

        global DATA
        global FRAMES
        if DATA is None:
            DATA = np.zeros((FRAMES, prob.shape[0], prob.shape[1]), dtype=prob.dtype)

        global FRAME_COUNT
        DATA[FRAME_COUNT] = prob
        FRAME_COUNT += 1

        global PROGRESS
        PROGRESS.update()

        if FRAME_COUNT == FRAMES:
            global DONE
            with DONE:
                DONE.notify()


def main() -> None:
    from docopt import docopt

    args = docopt(__doc__, version='1.0.0', options_first=True)

    import signal

    import h5py
    import paho.mqtt.client as mqtt
    from tqdm import tqdm

    machine = args['--machine']

    global FRAMES
    FRAMES = int(args['--frames'])

    file = args['FILE']

    signal.signal(signal.SIGINT, shutdown)
    signal.signal(signal.SIGTERM, shutdown)

    global CLIENT
    client = mqtt.Client(client_id=CLIENT)
    client.on_message = on_message
    client.connect(host=machine)
    client.loop_start()
    global TOPIC
    client.subscribe(topic=TOPIC)

    global PROGRESS
    PROGRESS = tqdm(total=FRAMES, desc=file)

    with DONE:
        DONE.wait()

    PROGRESS.close()

    client.unsubscribe(topic=TOPIC)
    client.loop_stop()
    client.disconnect()

    global DATA
    with h5py.File(file, 'w') as f:
        f.create_dataset(name='prob', data=DATA)

    print(f'Wrote {file}')


if __name__ == '__main__':
    main()



# SonusAI metrics utilities for model training and validation
from .calc_class_weights import calc_class_weights_from_mixdb
from .calc_class_weights import calc_class_weights_from_truth
from .calc_optimal_thresholds import calc_optimal_thresholds
from .calc_pcm import calc_pcm
from .calc_pesq import calc_pesq
from .calc_sa_sdr import calc_sa_sdr
from .calc_sample_weights import calc_sample_weights
from .calc_wer import calc_wer
from .calc_wsdr import calc_wsdr
from .class_summary import class_summary
from .confusion_matrix_summary import confusion_matrix_summary
from .one_hot import one_hot
from .snr_summary import snr_summary



from sox import Transformer as SoxTransformer

from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import ImpulseResponseData


def read_impulse_response(name: str) -> ImpulseResponseData:
    """Read impulse response data using SoX

    :param name: File name
    :return: ImpulseResponseData object
    """
    import numpy as np
    from scipy.io import wavfile

    from sonusai import SonusAIError
    from .datatypes import ImpulseResponseData
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    # Read impulse response data from audio file
    try:
        sample_rate, data = wavfile.read(expanded_name)
    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')

    data = data.astype(np.float32)
    offset = np.argmax(data)
    data = data[offset:]
    data = data / np.linalg.norm(data)

    return ImpulseResponseData(name=name, sample_rate=sample_rate, data=data)


def read_audio(name: str) -> AudioT:
    """Read audio data from a file using SoX

    :param name: File name
    :return: Array of time domain audio data
    """
    from typing import Any

    import numpy as np
    from sox.core import sox

    from sonusai import SonusAIError
    from .constants import BIT_DEPTH
    from .constants import CHANNEL_COUNT
    from .constants import ENCODING
    from .constants import SAMPLE_RATE
    from .tokenized_shell_vars import tokenized_expand

    def encode_output(buffer: Any) -> np.ndarray:
        from sonusai import SonusAIError
        from .constants import BIT_DEPTH
        from .constants import ENCODING

        if BIT_DEPTH == 8:
            return np.frombuffer(buffer, dtype=np.int8)

        if BIT_DEPTH == 16:
            return np.frombuffer(buffer, dtype=np.int16)

        if BIT_DEPTH == 24:
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 32:
            if ENCODING == 'floating-point':
                return np.frombuffer(buffer, dtype=np.float32)
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 64:
            return np.frombuffer(buffer, dtype=np.float64)

        raise SonusAIError(f'Invalid BIT_DEPTH {BIT_DEPTH}')

    expanded_name, _ = tokenized_expand(name)

    try:
        # Read in and convert to desired format
        # NOTE: pysox format transformations do not handle encoding properly; need to use direct call to sox instead
        args = [
            '-D',
            '-G',
            expanded_name,
            '-t', 'raw',
            '-r', str(SAMPLE_RATE),
            '-b', str(BIT_DEPTH),
            '-c', str(CHANNEL_COUNT),
            '-e', ENCODING,
            '-',
            'remix', '1',
        ]
        status, out, err = sox(args, None, False)
        if status != 0:
            raise SonusAIError(f'sox stdout: {out}\nsox stderr: {err}')

        return encode_output(out)

    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}):\n{e}')
        else:
            raise SonusAIError(f'Error reading {name}:\n{e}')


class Transformer(SoxTransformer):
    """Override certain sox.Transformer methods
    """

    def fir(self, coefficients):
        """Use SoX’s FFT convolution engine with given FIR filter coefficients.

        The SonusAI override allows coefficients to be either a list of numbers
        or a string containing a text file with the coefficients.

        Parameters
        ----------
        coefficients : list or str
            fir filter coefficients

        """
        from sox.core import is_number

        from sonusai import SonusAIError

        if not isinstance(coefficients, list) and not isinstance(coefficients, str):
            raise SonusAIError("coefficients must be a list or a str.")

        if isinstance(coefficients, list) and not all([is_number(c) for c in coefficients]):
            raise SonusAIError("coefficients list must be numbers.")

        effect_args = ['fir']
        if isinstance(coefficients, list):
            effect_args.extend(['{:f}'.format(c) for c in coefficients])
        else:
            effect_args.append(coefficients)

        self.effects.extend(effect_args)
        self.effects_log.append('fir')

        return self

    def tempo(self, factor, audio_type=None, quick=False):
        """Time stretch audio without changing pitch.

        This effect uses the WSOLA algorithm. The audio is chopped up into
        segments which are then shifted in the time domain and overlapped
        (cross-faded) at points where their waveforms are most similar as
        determined by measurement of least squares.

        The SonusAI override does not generate a warning for small factors.
        The sox.Transformer's implementation of stretch does not invert
        the factor even though it says that it does; this invalidates the
        factor size check and produces the wrong result.

        Parameters
        ----------
        factor : float
            The ratio of new tempo to the old tempo.
            For ex. 1.1 speeds up the tempo by 10%; 0.9 slows it down by 10%.
        audio_type : str
            Type of audio, which optimizes algorithm parameters. One of:
             * m : Music,
             * s : Speech,
             * l : Linear (useful when factor is close to 1),
        quick : bool, default=False
            If True, this effect will run faster but with lower sound quality.

        See Also
        --------
        stretch, speed, pitch

        """
        from sox.core import is_number
        from sox.log import logger

        from sonusai import SonusAIError

        if not is_number(factor) or factor <= 0:
            raise SonusAIError('factor must be a positive number')

        if factor < 0.5 or factor > 2:
            logger.warning('Using an extreme time stretching factor. Quality of results will be poor')

        if audio_type not in [None, 'm', 's', 'l']:
            raise SonusAIError("audio_type must be one of None, 'm', 's', or 'l'.")

        if not isinstance(quick, bool):
            raise SonusAIError('quick must be a boolean')

        effect_args = ['tempo']

        if quick:
            effect_args.append('-q')

        if audio_type is not None:
            effect_args.append('-{}'.format(audio_type))

        effect_args.append('{:f}'.format(factor))

        self.effects.extend(effect_args)
        self.effects_log.append('tempo')

        return self

    def build_array(self, input_filepath=None, input_array=None,
                    sample_rate_in=None, extra_args=None):
        """Given an input file or array, returns the output as a numpy array
        by executing the current set of commands. By default, the array will
        have the same sample rate as the input file unless otherwise specified
        using set_output_format. Functions such as channels and convert
        will be ignored!

        The SonusAI override does not generate a warning for rate transforms.

        Parameters
        ----------
        input_filepath : str or None
            Either path to input audio file or None.
        input_array : np.ndarray or None
            A np.ndarray of a waveform with shape (n_samples, n_channels).
            If this argument is passed, sample_rate_in must also be provided.
            If None, input_filepath must be specified.
        sample_rate_in : int
            Sample rate of input_array.
            This argument is ignored if input_array is None.
        extra_args : list or None, default=None
            If a list is given, these additional arguments are passed to SoX
            at the end of the list of effects.
            Don't use this argument unless you know exactly what you're doing!

        Returns
        -------
        output_array : np.ndarray
            Output audio as a numpy array

        Examples
        --------

        > import numpy as np
        > import sox
        > tfm = sox.Transformer()
        > sample_rate = 44100
        > y = np.sin(2 * np.pi * 440.0 * np.arange(sample_rate * 1.0) / sample_rate)

        file in, array out

        > output_array = tfm.build(input_filepath='path/to/input.wav')

        array in, array out

        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)

        specifying the output sample rate

        > tfm.set_output_format(rate=8000)
        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)

        if an effect changes the number of channels, you must explicitly
        specify the number of output channels

        > tfm.remix(remix_dictionary={1: [1], 2: [1], 3: [1]})
        > tfm.set_output_format(channels=3)
        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)


        """
        import numpy as np

        from sox.core import SoxError
        from sox.core import sox
        from sox.log import logger
        from sox.transform import ENCODINGS_MAPPING

        input_format, input_filepath = self._parse_inputs(
            input_filepath, input_array, sample_rate_in
        )

        # check if any of the below commands are part of the effects chain
        ignored_commands = ['channels', 'convert']
        if set(ignored_commands) & set(self.effects_log):
            logger.warning(
                "When outputting to an array, channels and convert " +
                "effects may be ignored. Use set_output_format() to " +
                "specify output formats."
            )

        output_filepath = '-'

        if input_format.get('file_type') is None:
            encoding_out = np.int16
        else:
            encoding_out = [
                k for k, v in ENCODINGS_MAPPING.items()
                if input_format['file_type'] == v
            ][0]

        n_bits = np.dtype(encoding_out).itemsize * 8

        output_format = {
            'file_type':       'raw',
            'rate':            sample_rate_in,
            'bits':            n_bits,
            'channels':        input_format['channels'],
            'encoding':        None,
            'comments':        None,
            'append_comments': True,
        }

        if self.output_format.get('rate') is not None:
            output_format['rate'] = self.output_format['rate']

        if self.output_format.get('channels') is not None:
            output_format['channels'] = self.output_format['channels']

        if self.output_format.get('bits') is not None:
            n_bits = self.output_format['bits']
            output_format['bits'] = n_bits

        match n_bits:
            case 8:
                encoding_out = np.int8
            case 16:
                encoding_out = np.int16
            case 32:
                encoding_out = np.float32
            case 64:
                encoding_out = np.float64
            case _:
                raise ValueError("invalid n_bits {}".format(n_bits))

        args = []
        args.extend(self.globals)
        args.extend(self._input_format_args(input_format))
        args.append(input_filepath)
        args.extend(self._output_format_args(output_format))
        args.append(output_filepath)
        args.extend(self.effects)

        if extra_args is not None:
            if not isinstance(extra_args, list):
                raise ValueError("extra_args must be a list.")
            args.extend(extra_args)

        status, out, err = sox(args, input_array, False)
        if status != 0:
            raise SoxError(
                "Stdout: {}\nStderr: {}".format(out, err)
            )

        out = np.frombuffer(out, dtype=encoding_out)
        if output_format['channels'] > 1:
            out = out.reshape(
                (
                    output_format['channels'],
                    int(len(out) / output_format['channels'])
                ), order='F'
            ).T
        logger.info(
            "Created array with effects: %s",
            " ".join(self.effects_log)
        )

        return out



import numpy as np


def calc_pcm(hypothesis: np.ndarray,
             reference: np.ndarray,
             with_log: bool = False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate phase constrained magnitude error

    These must include a noise to make a complete mixture estimate, i.e.,
        noise_est = mixture - sum-over-nsrc(s_est(:, nsrc, :))
    should be one of the sources in s_true and s_est.

    Calculates mean-over-srcs(mean-over-tf(| (|Sr(t, f)| + |Si(t, f)|) - (|Shr(t, f)| + |Shi(t, f)|) |))

    Reference:
        Self-attending RNN for Speech Enhancement to Improve Cross-corpus Generalization
        Ashutosh Pandey, Student Member, IEEE and DeLiang Wang, Fellow, IEEE
        https://doi.org/10.48550/arXiv.2105.12831

    :param hypothesis: complex [frames, nsrc, bins]
    :param reference: complex [frames, nsrc, bins]
    :param with_log: enable log
    :return: (error, error per bin, error per frame)
    """
    # LSM = 1/(T*F) * sumtf(| (|Sr(t, f)| + |Si(t, f)|) - (|Shr(t, f)| + |Shi(t, f)|) |)
    # LPCM = 1/2 * LSM(s, sh) + 1/2 * LSM(n, nh)

    # [frames, nsrc, bins]
    hypothesis_abs = np.abs(np.real(hypothesis)) + np.abs(np.imag(hypothesis))
    reference_abs = np.abs(np.real(reference)) + np.abs(np.imag(reference))
    err = np.abs(reference_abs - hypothesis_abs)

    # mean over frames, nsrc for value per bin
    err_b = np.mean(np.mean(err, axis=0), axis=0)
    # mean over bins, nsrc for value per frame
    err_f = np.mean(np.mean(err, axis=2), axis=1)
    # mean over bins and frames, nsrc for scalar value
    err = np.mean(np.mean(err, axis=(0, 2)), axis=0)

    if with_log:
        err_b = np.around(20 * np.log10(err_b + np.finfo(np.float32).eps), 3)
        err_f = np.around(20 * np.log10(err_f + np.finfo(np.float32).eps), 3)
        err = np.around(20 * np.log10(err + np.finfo(np.float32).eps), 3)

    return err, err_b, err_f
